/******************************************************************************
  ShiftBriteDriver.c  Driving a ShiftBrite RGB LED module (Allegro IC A6281)
  Copyright; Sep 2012 Roman Black
  Open source; use it for whatever you like, but please mention me as the
  original author and mention my web page; www.RomanBlack.com

  PIC 16F628A - 8MHz HS xtal
  Description; This is a very simple serial driver to send the colour
  and brightness data to a ShiftBrite RGB LED module. Multiple ShiftBrites
  can be chained for large RGB matrix displays, using this same code.
  
  Using; There are two functions to set the ShiftBrite RGB. One uses
  simpler 8bit variables and gives lower resolution control of each
  LED so 0-255 becomes 0-1020. The other function requires 16bit variables,
  and gives full resolution of each colour LED 0-1023. Use the function
  you prefer. Another config function sets the colour balance of the
  three LEDs and allows max brightness from the ShiftBrite module.
  
  Connection; The ShiftBrite need 4 control wires; Data, Clock, Latch
  and OutputEnable (DI, CI, LI and EI). If LED sleep is not needed you can
  just tie OutputEnable always ON (low) like I did in this example.

  Compatibility; should be OK on any PIC in C, also suited for very small PICs.
******************************************************************************/
// Set PIC port pins to drive the ShiftBrite (its EI pin is just tied low);
#define SHIFT_TRIS    TRISB   // TRISx
#define SHIFT_PORT    PORTB   // use PORTx for PIC16, use LATx for PIC18
#define PIN_DAT       F0      // DI
#define PIN_CLK       F1      // CI
#define PIN_LAT       F2      // LI

// Private RAM vars used by ShiftBrite functions;
unsigned long string32  absolute 0x20;     // the 32bit string to send
unsigned char string32a absolute 0x20;     // (overloaded for faster byte access)
unsigned char string32b absolute 0x21;
unsigned char string32c absolute 0x22;
unsigned char string32d absolute 0x23;
unsigned long temp32;       // for building 32bit data

// private ShiftBrite functions;
void send_ShiftBrite32bits(unsigned char latch);

// public (User) functions;
void set_ShiftBriteRGBL8(unsigned char red, unsigned char green,
                          unsigned char blue, unsigned char latch); // 8bit data
void set_ShiftBriteRGBL10(unsigned int red, unsigned int green,
                          unsigned int blue, unsigned char latch); // 10bit data
void set_ShiftBriteConfig(unsigned char redb, unsigned char greenb,
                          unsigned char blueb, unsigned char latch); // set Config

// User vars for my test code in main()
unsigned int r;
unsigned int g;
unsigned int b;
unsigned int i;

//=============================================================================
//  send_ShiftBrite32bits
//=============================================================================
void send_ShiftBrite32bits(unsigned char latch)
{
  //-------------------------------------------------------
  // The data to send is already in 32bit variable shift32.
  //-------------------------------------------------------
  unsigned char port_shad;
  unsigned char si;
  // send the 32bit string to the ShiftBrite(s), MSB first
  for(si=0; si<32; si++)
  {
    // set data bit first, all bits using shadow register for reliability
    port_shad = 0;
    if(string32d.F7) port_shad.PIN_DAT = 1;
    SHIFT_PORT = port_shad;
    asm NOP;    // tiny safe delay
    asm NOP;
    // make a clock bit /  edge
    port_shad.PIN_CLK = 1;
    SHIFT_PORT = port_shad;
    // shift data to get the next bit
    string32 <<= 1;
  }
  // if latch is selected then also make a latch bit to write the new colours
  if(latch)
  {
    port_shad.PIN_LAT = 1;
    SHIFT_PORT = port_shad;
    asm NOP;
    asm NOP;
  }
  SHIFT_PORT = 0;     // clear all output lines when done
  asm NOP;
  asm NOP;
}
//-----------------------------------------------------------------------------


//=============================================================================
//  set_ShiftBriteConfig
//=============================================================================
void set_ShiftBriteConfig(unsigned char redb, unsigned char greenb,
                          unsigned char blueb, unsigned char latch)
{
  //-------------------------------------------------------
  // The A6281 IC can be set for different colour balance for
  // each LED. I found on my ShiftBrite module the blue LED was not
  // quite as bright. This function allows the config setting of
  // the 3 colour balance registers (7bits each; 0-127).
  // The ShiftBrite powers up at 36% brightness and for full brightness
  // you must first set all the colour balance registers to 127.
  // (From Allegro A6281 datasheet, 32 bits);
  // x M ---BBBBBBB ---RRRRRRR ---GGGGGGG (M is Mode, is 1 for Config)
  // - - ---M-----L ---M-----L ---M-----L (MSB--LSB)
  //-------------------------------------------------------
  // Compile the 3 lots of 8bit RGB data into 10bit spaces within 32bit string
  // this is not optimised for speed, but it works ok.
  temp32 = (blueb & 0b01111111);
  temp32 = (temp32 * 1048576);        // data <<20 bits
  string32 = temp32;
  temp32 = (redb & 0b01111111);
  temp32 = (temp32 * 1024);           // data <<10 bits
  string32 += temp32;
  string32 += (greenb & 0b01111111);  // no shift
  string32d.F6 = 1;     // set Mode to config
  send_ShiftBrite32bits(latch);
}
//-----------------------------------------------------------------------------


//=============================================================================
//  set_ShiftBriteRGBL8    Function for 8bit RGB control with 8bit variables
//=============================================================================
void set_ShiftBriteRGBL8(unsigned char red, unsigned char green,
                          unsigned char blue, unsigned char latch)
{
  //-------------------------------------------------------
  // The A6281 IC in the ShiftBrite needs 10bit RGB data 0-1023
  // so we must convert the vars 0-255 to 0-1020.
  // Then the brightness data is just shifted out the DAT and CLK
  // pins as one 32bit control string;
  // (From Allegro A6281 datasheet, 32 bits);
  // x M BBBBBBBBBB RRRRRRRRRR GGGGGGGGGG (M is Mode, is 0 for RGB set)
  // - - M--------L M--------L M--------L (MSB--LSB)
  // - - BBBBBBBB-- RRRRRRRR-- GGGGGGGG-- (we insert 8bit data like this)
  //-------------------------------------------------------
  // Compile the 3 lots of 8bit RGB data into 10bit spaces within 32bit string
  // this is not optimised for speed, but it works ok.
  temp32 = blue;
  temp32 = (temp32 * 4194304);    // data <<22 bits
  string32 = temp32;
  temp32 = red;
  temp32 = (temp32 * 4096);       // data <<12 bits
  string32 += temp32;
  temp32 = green;
  temp32 = (temp32 * 4);          // data <<2 bits
  string32 += temp32;
  send_ShiftBrite32bits(latch);
}
//-----------------------------------------------------------------------------


//=============================================================================
//  set_ShiftBriteRGBL10    Function for 10bit RGB control with 16bit variables
//=============================================================================
void set_ShiftBriteRGBL10(unsigned int red, unsigned int green,
                          unsigned int blue, unsigned char latch)
{
  //-------------------------------------------------------
  // The A6281 IC in the ShiftBrite needs 10bit RGB data 0-1023.
  // Then the brightness data is just shifted out the DAT and CLK
  // pins as one 32bit control string;
  // (From Allegro A6281 datasheet, 32 bits, shown MSbit first);
  // x M BBBBBBBBBB RRRRRRRRRR GGGGGGGGGG (M is Mode, should be 0)
  // - - M--------L M--------L M--------L (MSB--LSB)
  //-------------------------------------------------------
  // Compile the 3 lots of 10bit RGB data into 10bit spaces within 32bit string
  // This is not optimised for speed, but it works ok.
  temp32 = blue;
  temp32 = (temp32 * 1048576);    // data <<20 bits
  string32 = temp32;
  temp32 = red;
  temp32 = (temp32 * 1024);       // data <<10 bits
  string32 += temp32;
  string32 += green;              // no shift
  send_ShiftBrite32bits(latch);
}
//-----------------------------------------------------------------------------


//=============================================================================
//  MAIN
//=============================================================================
void main(void)
{
  //-------------------------------------------------------
  // PIC 16F628A setup
  CMCON = 0x07;   // comparators OFF
  PORTA = 0b00000000;
  TRISA = 0b00000000;       // PORTA all outputs (safe)

  SHIFT_PORT = 0;           // PORT used for ShiftBrite output pins
  SHIFT_TRIS = 0b00000000;  // (all outs) check for your wiring

  //-------------------------------------------------------
  // Must first set the ShiftBrite config  colour balances registers.
  // This is also needed to enable the full range of brightness from
  // the ShiftBrite module. 127 is max for each colour balance.
  set_ShiftBriteConfig(20,35,127,1);  // RGB balance, my blue LED needed lots more drive

  // The rest of this code is just a crude example of fading RGB colours
  r = 0;
  g = 0;
  b = 0;
  while(1)
  {
    // simple 3 phase colour ramp code, each ramp is +/-1000 steps (2000 total)
    // so this gives 3 phases of 2000 steps
    for(i=0; i<6000; i++)   // 3 phases total of 1500 each
    {
      // red, phased RR----
      if(i < 1000) r++;
      else if(r > 0) r--;

      // green, phased --GG--
      if(i > 2000 && i < 3000) g++;
      else if(g > 0) g--;

      // blue, phased ----BB
      if(i > 4000 && i < 5000) b++;
      else if(b > 0) b--;

      // send the adjusted RGB values to the ShiftBrite module
      set_ShiftBriteRGBL10(r,g,b,1);   // using the 10bit resolution
      Delay_uS(200);                   // and a delay sets the fade time
    }
  }
}
//-----------------------------------------------------------------------------




